@@ -1,4 +1,5 @@
-.First.lib <- function(lib, pkg) {
+.onLoad <- function(lib, pkg) {
+    require("methods")
 
     where <- match(paste("package:", pkg, sep = ""), search())
     ver <- read.dcf(file.path(lib, pkg, "DESCRIPTION"), "Version")
@@ -6,6 +7,7 @@
     title <- read.dcf(file.path(lib, pkg, "DESCRIPTION"), "Title")
     title <- as.character(title)
     cat(paste(title, " (version ", ver, ")\n", sep = ""))
-    
-    library.dynam("rrcov", pkg, lib)
 }
+
+
+ 
